'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = build;

require('babel-polyfill');

/**
 * # Closure Eliminator
 */

var $boundArrowFunction = Symbol('boundArrowFunction');
var $usedEval = Symbol('usedEval');

function _ref2(_ref) {
  var node = _ref.node;
  return node._generated || node._compact;
}

function _ref(path) {
  return path.isFunction();
}

function _ref3(path) {
  return path.type !== 'ArrowFunctionExpression';
}

function _ref4(parentArrow) {
  parentArrow.node[$boundArrowFunction] = true;
}

function _enter(path) {
  var parentFunctions = path.getAncestry().filter(_ref),
      nearestNoArrowFunction = parentFunctions.findIndex(_ref3);
  parentFunctions.slice(0, nearestNoArrowFunction).forEach(_ref4);
}

function _ref5(path) {
  return path.isFunction();
}

function _ref6(parentArrow) {
  parentArrow.node[$usedEval] = true;
}

function _enter2(path) {
  if (path.node.name === 'eval' && path.parentPath.type === 'CallExpression') {
    path.getAncestry().filter(_ref5).forEach(_ref6);
  }
}

function _ref9(_ref8) {
  var path = _ref8.path;
  return !path.isProgram() || path.node.sourceType === 'module';
}

function _getAllParentScopes(scope) {
  var scopes = [];
  while (scope = scope.parent) {
    scopes.push(scope);
  }
  return scopes;
}

function _getAttachmentPositio(bestParent, prevPath) {
  var possibleSiblings = void 0;
  if (bestParent.isFunction()) {
    possibleSiblings = bestParent.get('body.body');
  } else if (bestParent.isProgram() || bestParent.isBlockStatement()) {
    possibleSiblings = bestParent.get('body');
  }
  var prevParents = prevPath.getAncestry();
  return possibleSiblings.find(function (x) {
    return prevParents.indexOf(x) !== -1;
  });
}

function _logAllProgram(path, label) {
  var rootNode = path.getAncestry().pop().node;
  console.error(label);
  console.error(require('babel-generator').default(rootNode).code);
  console.error('\n=======================================================\n');
}
function build(babel) {
  var t = babel.types;


  return {
    visitor: {
      Function: {
        exit: function exit(path) {
          var node = path.node;

          if (path.node._hoisted) {
            return;
          }
          if (path.isClassMethod() || path.isObjectMethod() || node[$boundArrowFunction] || node[$usedEval]) {
            return;
          }
          if (path.findParent(_ref2)) {
            path.skip();
            return;
          }
          var bestParentScope = getHighestCompatibleHoistedScope(path);
          if (bestParentScope) {
            var attachPath = _getAttachmentPositio(bestParentScope.path, path);
            // _logAllProgram(path, 'before');// debug
            moveToNewPosition(path, attachPath);
            // _logAllProgram(path, 'after');// debug
          }
        }
      },
      ThisExpression: {
        enter: _enter
      },
      Identifier: {
        enter: _enter2
      }
    }
  };

  function getHighestCompatibleHoistedScope(path) {
    path.scope.crawl(); // sibling plugins may not update scope of auto-generated functions
    var parentScopes = _getAllParentScopes(path.scope),
        parentBindings = path.scope.parent.getAllBindings();

    function _ref7(subPath) {
      return subPath.getAncestry().indexOf(path) !== -1;
    }

    for (var id in parentBindings) {
      var hasUsageOfBinding = [].concat(parentBindings[id].referencePaths).concat(parentBindings[id].constantViolations).some(_ref7);
      if (hasUsageOfBinding) {
        var idx = parentScopes.indexOf(parentBindings[id].scope);
        if (idx !== -1) {
          parentScopes.splice(idx + 1, Infinity);
        }
      }
    }
    return parentScopes.filter(_ref9).filter(function (scope) {
      return scope !== path.scope.parent;
    }).pop();
  }function moveToNewPosition(path, attachPath) {
    var node = path.node,
        scope = path.scope,
        newScope = attachPath.parentPath.scope;

    if (node.type === 'FunctionDeclaration') {
      var uid = newScope.generateUidIdentifierBasedOnNode(node.id);
      scope.rename(node.id.name, uid.name);
      scope.moveBindingTo(uid.name, newScope);
      node._hoisted = true;
      attachPath.insertBefore([node]);
      path.remove();
    } else {
      var _uid = path.parentPath.scope.generateUidIdentifierBasedOnNode(node.id),
          replacement = t.functionDeclaration(_uid, node.params, normalizeFunctionBody(node.body));
      if (node.id && node.id.name) {
        scope.rename(node.id.name, _uid.name);
      }
      replacement.loc = node.loc;
      replacement.generator = node.generator;
      replacement.async = node.async;
      replacement._hoisted = true;
      attachPath.insertBefore([replacement]);
      path.replaceWith(t.identifier(_uid.name));
      newScope.crawl();
    }
  }

  /**
   * Normalize a function body so that it is always a BlockStatement.
   */
  function normalizeFunctionBody(node) {
    if (node.type !== 'BlockStatement') {
      return t.blockStatement([t.returnStatement(node)]);
    } else {
      return node;
    }
  }
}